import logging
import click
from os.path import isfile
from geoseeq.cli.utils import use_common_state
from geoseeq.blob_constructors import resolve_brn
from .vc_dir import VCDir
from .clone import (
    clone_project,
    clone_sample,
)

logger = logging.getLogger('geoseeq_api')


@click.group('vc')
def cli_vc():
    """Experimental version control features."""
    pass


@cli_vc.command('clone')
@use_common_state
@click.option('--target-dir', default=".", help='Choose a target directory for stub files')
@click.option('--uuids/--names', default=False, help='Use UUIDs instead of names for objects')
@click.argument('brn')
def cli_vc_clone(state, target_dir, uuids, brn):
    """Clone GeoSeeq stub files from a project or sample to local storage.
    
    Only downloads stub files, not the files they link to.
    """
    knex = state.get_knex()
    object_type, obj = resolve_brn(knex, brn)
    if object_type == 'project':
        return clone_project(obj, target_dir, uuid=uuids)
    if object_type == 'sample':
        return clone_sample(obj, target_dir, uuid=uuids)
    pass


@cli_vc.command('download')
@use_common_state
@click.option('--extension', default='.gvcf', help='File extension for GeoSeeq version control files')
@click.argument('paths', nargs=-1)
def cli_vc_download(state, extension, paths):
    """Download files from GeoSeeq to local storage."""
    knex = state.get_knex()
    if len(paths) == 0: paths = ['.']
    for path in paths:
        for stub in VCDir(path, extension=extension).stubs():
            stub.download(knex)


@cli_vc.command('status')
@use_common_state
@click.option('--extension', default='.gvcf', help='File extension for GeoSeeq version control files')
@click.argument('paths', nargs=-1)
def cli_vc_status(state, extension, paths):
    """Check the status of all link files in the current folder or in specified paths. Recursive."""
    if len(paths) == 0: paths = ['.']
    for path in paths:
        for stub in VCDir(path, extension=extension).stubs():
            verified = 'checksum_matches' if stub.verify() else 'no_checksum_match'
            color = 'green' if verified == 'checksum_matches' else 'red'
            click.echo(click.style(f'{stub.brn}\t{stub.local_path}\t{verified}', fg=color))


@cli_vc.command('list')
@click.option('--outfile', type=click.File('w'), default='-', help='File to write paths to')
@click.option('--extension', default='.gvcf', help='File extension for GeoSeeq version control files')
@click.argument('paths', nargs=-1)
def cli_vc_list(outfile, extension, paths):
    """Recursively list all filepaths for files pointed to by stub files."""
    if len(paths) == 0: paths = ['.']
    for path in paths:
        for stub in VCDir(path, extension=extension).stubs():
            if isfile(stub.local_path):
                print(stub.local_path, file=outfile)


import click

from .. import Organization
from .utils import use_common_state


@click.group('add')
def cli_add():
    """Add samples to projects."""
    pass


@cli_add.command('samples-to-group')
@use_common_state
@click.option('--sample-manifest', type=click.File('r'),
              help='List of sample names to download from')
@click.argument('source_org_name')
@click.argument('source_grp_name')
@click.argument('destination_org_name')
@click.argument('destination_grp_name')
@click.argument('sample_names', nargs=-1)
def cli_add_samples(state, sample_manifest,
                    source_org_name, source_grp_name,
                    destination_org_name, destination_grp_name,
                    sample_names):
    """Add samples from the source group to the destination group."""
    knex = state.get_knex()
    source_org = Organization(knex, source_org_name).get()
    source_grp = source_org.sample_group(source_grp_name).get()
    dest_org = Organization(knex, destination_org_name).get()
    dest_grp = dest_org.sample_group(destination_grp_name).get()
    if sample_manifest:
        sample_names = set(sample_names) | set([el.strip() for el in sample_manifest if el])
    for sname in sample_names:
        sample = source_grp.sample(sname).get()
        dest_grp.add_sample(sample)
    dest_grp.save()
    click.echo(f'Added {len(sample_names)} to {destination_grp_name}')




import json
from .vc_stub import VCStub
from glob import glob


class VCDir:
    """A class to represent a directory with GeoSeeq Version Control files."""

    def __init__(self, root_path, extension='.gvcf'):
        self.path = root_path
        self.ext = extension

    def stubs(self):
        """Return an iterator over stub files."""
        for path in glob(f'{self.path}/**/*{self.ext}', recursive=True):
            yield VCStub.from_path(path)



import logging

from requests.exceptions import HTTPError

from .file_system_cache import FileSystemCache

logger = logging.getLogger("geoseeq_api")  # Same name as calling module
logger.addHandler(logging.NullHandler())  # No output unless configured by calling program


class RemoteObjectError(Exception):
    pass


class RemoteObjectOverwriteError(RemoteObjectError):
    pass


class RemoteObject:
    optional_remote_fields = []

    def __init__(self, *args, **kwargs):
        self._already_fetched = False
        self._modified = False
        self._deleted = False
        self.blob = None
        self.uuid = None
        self.cache = FileSystemCache()
        self.url_options = {}

    def __setattr__(self, key, val):
        if hasattr(self, "deleted") and self._deleted:
            logger.error(f"Attribute cannot be set, RemoteObject has been deleted. {self}")
            raise RemoteObjectError("This object has been deleted.")
        super(RemoteObject, self).__setattr__(key, val)
        if key in self.remote_fields or key == self.parent_field:
            logger.debug(f'Setting RemoteObject modified. key "{key}"')
            super(RemoteObject, self).__setattr__("_modified", True)

    @property
    def inherited_url_options(self):
        opts = self.url_options.copy()
        if self.parent_field:
            parent = getattr(self, self.parent_field)
            opts.update(parent.inherited_url_options)
        return opts

    def invalidate_cache(self):
        self.cache.clear_blob(self)

    def get_cached_blob(self):
        return self.cache.get_cached_blob(self)

    def cache_blob(self, blob):
        return self.cache.cache_blob(self, blob)

    def load_blob(self, blob, allow_overwrite=False):
        logger.debug(f"Loading blob. {blob}")
        if self._deleted:
            logger.error(f"Cannot load blob, RemoteObject has been deleted. {self}")
            raise RemoteObjectError("This object has been deleted.")
        for field in self.remote_fields:
            current = getattr(self, field, None)
            try:
                new = blob[field]
            except KeyError:
                if field not in self.optional_remote_fields:
                    logger.error(f"Blob being loaded is missing key. {field}")
                    raise KeyError(
                        f"Key {field} is missing for object {self} (type {type(self)})\
                             in blob: {blob}"
                    )
                new = None
            if not allow_overwrite and current and current != new:
                is_overwrite = True
                if isinstance(current, dict) and isinstance(new, dict):
                    append_only = True
                    for k, v in current.items():
                        if (k not in new) or (new[k] != v):
                            append_only = False
                        break
                    if append_only:
                        is_overwrite = False
                if is_overwrite:
                    logger.error(f"Loading blob would overwrite key. {field}")
                    raise RemoteObjectOverwriteError(
                        (
                            f'Loading blob would overwrite field "{field}":\n\t'
                            f'current: "{current}" (type: "{type(current)}")\n\t'
                            f'new:     "{new}" (type: "{type(new)}")'
                        )
                    )
            setattr(self, field, new)

    def exists(self, allow_overwrite=False):
        try:
            self.get(allow_overwrite=allow_overwrite)
            return True
        except HTTPError:
            return False

    def get(self, allow_overwrite=False):
        """Fetch the object from the server."""
        if self._deleted:
            logger.error(f"Cannot GET blob, RemoteObject has been deleted. {self}")
            raise RemoteObjectError("This object has been deleted.")
        if not self._already_fetched:
            logger.debug(f"Fetching RemoteBlob. {self}")
            self._get(allow_overwrite=allow_overwrite)
            self._already_fetched = True
            self._modified = False
        else:
            logger.debug(f"RemoteObject has already been fetched. {self}")
        return self

    def create(self):
        """Create this object on the server."""
        if self._deleted:
            logger.error(f"Cannot create blob, RemoteObject has been deleted. {self}")
            raise RemoteObjectError("This object has been deleted.")
        if not self._already_fetched:
            logger.debug(f"Creating RemoteBlob. {self}")
            self.cache.clear_blob(self)
            self._create()
            self._already_fetched = True
            self._modified = False
        else:
            logger.debug(f"RemoteObject has already been fetched. {self}")
        return self

    def save(self):
        """Assuming the object exists on the server make the server-side object
        match the state of this object.
        """
        if self._deleted:
            logger.error(f"Cannot save blob, RemoteObject has been deleted. {self}")
            raise RemoteObjectError("This object has been deleted.")
        if not self._already_fetched:
            msg = "Attempting to SAVE an object which has not been fetched is disallowed."
            raise RemoteObjectError(msg)
        if self._modified:
            logger.debug(f"Saving RemoteBlob. {self}")
            self.cache.clear_blob(self)
            self._save()
            self._modified = False
        else:
            logger.debug(f"RemoteBlob has not been modified. Nothing to save. {self}")

    def idem(self):
        """Make the state of this object match the server."""
        if self._deleted:
            raise RemoteObjectError("This object has been deleted.")
        if not self._already_fetched:
            try:
                self.get()
            except HTTPError:
                self.create()
        else:
            self.save()
        return self

    def delete(self):
        logger.debug(f"Deleting RemoteBlob. {self}")
        self.knex.delete(self.nested_url())
        self._already_fetched = False
        self._deleted = True

    @classmethod
    def all_uuids(self, knex):
        """Return a list of all objects of this type."""
        results = knex.get(self.url_prefix)['results']
        return [result['uuid'] for result in results]






"""Test suite for experimental functions."""
import random
from os import environ
from unittest import TestCase

ENDPOINT = environ.get("GEOSEEQ_API_TESTING_ENDPOINT", "http://127.0.0.1:8000")


def random_str(len=12):
    """Return a random alphanumeric string of length `len`."""
    out = random.choices("abcdefghijklmnopqrtuvwxyzABCDEFGHIJKLMNOPQRTUVWXYZ0123456789", k=len)
    return "".join(out)


class TestGeoseeqApiWorkOrders(TestCase):
    """Test suite for packet building."""

    pass



import click

dryrun_option = click.option('--dryrun/--wetrun', default=False, help='Print what will be created without actually creating it')
overwrite_option = click.option('--overwrite/--no-overwrite', default=False, help='Overwrite existing samples, files, and data')
module_option = lambda x: click.option('-m', '--module-name', type=click.Choice(x), default=x[0], help='Name for the module that will store the data')
private_option = click.option('--private/--public', default=True, help='Make objects private (default) or public')
link_option = click.option(
    '--link-type',
    default='upload',
    type=click.Choice(['upload', 's3', 'ftp', 'azure', 'sra']),
    help='Link the files from a cloud storage service instead of copying them'
)
yes_option = click.option('--yes/--confirm', default=False, help='Skip confirmation prompts')
sample_manifest_option = click.option(
    "--sample-manifest", type=click.File("r"), help="List of sample names to download from"
)
org_arg = click.argument('org_name')
project_arg = click.argument('project_name')
sample_arg = click.argument('sample_name')
module_arg = click.argument('module_name')
field_name = click.argument('field_name')
project_id_arg = click.argument('project_id', nargs=1)
sample_ids_arg = click.argument('sample_ids', nargs=-1)
folder_id_arg = click.argument('folder_id', nargs=1)


import json
import logging
import os
from glob import glob
from hashlib import sha256
from random import randint
from time import time

logger = logging.getLogger(__name__)  # Same name as calling module
logger.addHandler(logging.NullHandler())  # No output unless configured by calling program
CACHED_BLOB_TIME = 3 * 60 * 60  # 3 hours in seconds
CACHE_DIR = os.environ.get('GEOSEEQ_API_CACHE_DIR', '.')


def hash_obj(obj):
    val = obj
    if not isinstance(obj, str):
        val = obj.pre_hash()
    result = sha256(val.encode())
    result = result.hexdigest()
    return result


def time_since_file_cached(blob_filepath):
    timestamp = int(blob_filepath.split('__')[-1].split('.json')[0])
    elapsed_time = int(time()) - timestamp
    return elapsed_time


class FileSystemCache:

    def __init__(self, timeout=CACHED_BLOB_TIME):
        self.no_cache = 'false' in os.environ.get('USE_GEOSEEQ_CACHE', 'TRUE').lower()
        self.timeout = timeout

    def clear_blob(self, obj):
        if self.no_cache:
            return
        blob_filepath, path_exists = self.get_cached_blob_filepath(obj)
        if path_exists:
            logger.debug(f'Clearing cached blob. {blob_filepath}')
            try:
                os.remove(blob_filepath)
            except FileNotFoundError:
                logger.debug(f'Blob was deleted before it could be removed. {obj}')
                pass

    def get_cached_blob_filepath(self, obj):
        path_base = f'{CACHE_DIR}/.geoseeq_api_cache/v1/geoseeq_api_cache__{hash_obj(obj)}'
        os.makedirs(os.path.dirname(path_base), exist_ok=True)
        paths = sorted(glob(f'{path_base}__*.json'))
        if paths:
            return paths[-1], True
        timestamp = int(time())
        blob_filepath = f'{path_base}__{timestamp}.json'
        return blob_filepath, False

    def get_cached_blob(self, obj):
        if self.no_cache:
            return None
        logger.debug(f'Getting cached blob. {obj}')
        blob_filepath, path_exists = self.get_cached_blob_filepath(obj)
        if not path_exists:  # cache not found
            logger.debug(f'No cached blob found. {obj}')
            return None
        elapsed_time = time_since_file_cached(blob_filepath)
        if elapsed_time > (self.timeout + randint(0, self.timeout // 10)):  # cache is stale
            logger.debug(f'Found stale cached blob. {obj}')
            os.remove(blob_filepath)
            return None
        logger.debug(f'Found good cached blob. {obj}')
        try:
            blob = json.loads(open(blob_filepath).read())
            return blob
        except FileNotFoundError:
            logger.debug(f'Blob was deleted before it could be returned. {obj}')
            return None

    def cache_blob(self, obj, blob):
        if self.no_cache:
            return None
        logger.debug(f'Caching blob. {obj} {blob}')
        blob_filepath, path_exists = self.get_cached_blob_filepath(obj)
        if path_exists:  # save a new cache if an old one exists
            elapsed_time = time_since_file_cached(blob_filepath)
            if elapsed_time < ((self.timeout / 2) + randint(0, self.timeout // 10)):
                # Only reload a file if it is old enough
                return
            self.clear_blob(obj)
            return self.cache_blob(obj, blob)
        with open(blob_filepath, 'w') as f:
            f.write(json.dumps(blob))



from .result import (
    SampleResultFolder,
    SampleResultFile,
    ProjectResultFolder,
    ProjectResultFile,
    SampleAnalysisResult,
    SampleAnalysisResultField,
    SampleGroupAnalysisResult,
    SampleGroupAnalysisResultField,
)
from .knex import (
    GeoseeqForbiddenError,
    GeoseeqGeneralError,
    GeoseeqInternalError,
    GeoseeqNotFoundError,
    GeoseeqOtherError,
    GeoseeqTimeoutError,
    Knex,
)
from .organization import Organization
from .pipeline import Pipeline, PipelineModule, PipelineRun
from .remote_object import RemoteObjectError, RemoteObjectOverwriteError
from .sample import Sample
from .project import Project, SampleGroup
from .user import User
from .work_orders import (
    GroupWorkOrder,
    GroupWorkOrderProto,
    JobOrder,
    WorkOrder,
    WorkOrderProto,
)



import json
import logging
import os
import time
import urllib.request
from os.path import basename, getsize, join
from pathlib import Path
from tempfile import NamedTemporaryFile

import requests

from geoseeq.constants import FIVE_MB
from geoseeq.remote_object import RemoteObject, RemoteObjectError
from geoseeq.utils import download_ftp, md5_checksum

logger = logging.getLogger("geoseeq_api")  # Same name as calling module
logger.addHandler(logging.NullHandler())  # No output unless configured by calling program


def _download_head(url, filename, head=None):
    if head and head > 0:
        opener = urllib.request.build_opener()
        if head:
            opener.addheaders = [('Range', f'bytes=0-{head}')]
        urllib.request.install_opener(opener)
    try:
        urllib.request.urlretrieve(url, filename)  # can throw 416 error if head is too large
    except urllib.error.HTTPError as e:
        if e.code == 416:
            logger.warning(f"HEAD request failed, trying again without HEAD.")
            _download_head(url, filename, head=None)
        else:
            raise e
    

def diff_dicts(blob1, blob2):
    for problem in _diff_dicts("original", "$", blob1, blob2):
        yield problem
    for problem in _diff_dicts("serialized", "$", blob2, blob1):
        yield problem


def _diff_lists(suffix, depth, l1, l2):
    if len(l1) != len(l2):
        yield (f"MISMATCHED_LENGTHS:{suffix}:{depth}", len(l1), len(l2))
    for i in range(len(l1)):
        v1, v2 = l1[i], l2[i]
        if v1 != v2:
            mydepth = f"{depth}.index_{i}"
            if not isinstance(v1, type(v2)):
                yield (f"MISMATCHED_TYPES:{suffix}:{mydepth}", v1, v2)
            elif isinstance(v1, dict):
                for problem in _diff_dicts(suffix, mydepth, v1, v2):
                    yield problem
            elif isinstance(v1, list):
                for problem in _diff_lists(suffix, mydepth, v1, v2):
                    yield problem
            else:
                yield (f"MISMATCHED_LIST_VALUES:{suffix}:{mydepth}", v1, v2)


def _diff_dicts(suffix, depth, blob1, blob2):
    for k, v in blob1.items():
        if k not in blob2:
            yield (f"MISSING_KEY:{suffix}:{depth}", k, None)
        elif v != blob2[k]:
            if not isinstance(v, type(blob2[k])):
                yield (f"MISMATCHED_TYPES:{suffix}:{depth}", v, blob2[k])
            elif isinstance(v, dict):
                for problem in _diff_dicts(suffix, depth + "." + k, v, blob2[k]):
                    yield problem
            elif isinstance(v, list):
                for problem in _diff_lists(suffix, depth + "." + k, v, blob2[k]):
                    yield problem
            else:
                yield (f"MISMATCHED_DICT_VALUES:{suffix}:{depth}", v, blob2[k])


def check_json_serialization(blob):
    """Raise an error if serialization+deserialization fails to return an exact copy."""
    json_serialized = json.loads(json.dumps(blob))
    if json_serialized == blob:
        return
    if isinstance(blob, dict):
        issues = list(diff_dicts(blob, json_serialized))
    else:
        issues = [("MISMATCHED_VALUES:values_only:0", blob, json_serialized)]
    issues = [str(el) for el in issues]
    issues = "\n".join(issues)
    raise RemoteObjectError(f"JSON Serialization modifies object\nIssues:\n{issues}")




import click

from .. import Organization, Project, Sample
from .utils import use_common_state
from ..utils import paginated_iterator


@click.group('list')
def cli_list():
    """List objects on GeoSeeq."""
    pass


@cli_list.command('samples')
@use_common_state
@click.argument('org_name')
@click.argument('grp_name')
def cli_list_samples(state, org_name, grp_name):
    """Print a list of samples in the specified group."""
    knex = state.get_knex()
    org = Organization(knex, org_name).get()
    grp = org.sample_group(grp_name).get()
    for sample in grp.get_samples():
        print(sample, file=state.outfile)


@cli_list.command('organizations')
@use_common_state
def cli_list_organizations(state):
    """Print a list of organizations."""
    knex = state.get_knex()
    for uuid in Organization.all_uuids(knex):
        print(uuid, file=state.outfile)


@cli_list.group('uuids')
def cli_list_uuids():
    """Print lists of UUIDs for objects on GeoSeeq."""
    pass


@cli_list_uuids.command('projects')
@use_common_state
def cli_list_projects(state):
    """Print a list of projects in the specified organization."""
    knex = state.get_knex()
    for uuid in Project.all_uuids(knex):
        print(uuid, file=state.outfile)


@cli_list_uuids.command('samples')
@use_common_state
def cli_list_samples(state):
    """Print a list of samples in the specified organization."""
    knex = state.get_knex()
    for proj_uuid in Project.all_uuids(knex):
        click.echo(f"Project: {proj_uuid}", err=True)
        url = f"sample_groups/{proj_uuid}/samples"
        for sample_blob in paginated_iterator(knex, url,):
            uuid = sample_blob['uuid']
            print(uuid, file=state.outfile)


import click

from .. import Organization
from ..blob_constructors import sample_ar_from_uuid, sample_from_uuid
from .constants import *
from .utils import use_common_state


@click.group('create')
def cli_create():
    """Create objects on Geoseeq."""
    pass


@cli_create.command('org')
@use_common_state
@click.argument('org_name')
def cli_create_org(state, org_name):
    """Create an organization on Geoseeq."""
    knex = state.get_knex()
    org = Organization(knex, org_name).create()
    click.echo(f'Created: {org}', err=True)


@cli_create.command('sample-group')
@use_common_state
@click.option('--private/--public', default=True)
@click.option('--library/--not-library', default=True)
@click.argument('org_name')
@click.argument('grp_name')
def cli_create_grp(state, private, library, org_name, grp_name):
    """Create a sample group on Geoseeq."""
    knex = state.get_knex()
    org = Organization(knex, org_name).get()
    grp = org.sample_group(grp_name, is_library=library, is_public=not private).create()
    click.echo(f'Created: {grp}', err=True)


def simplify_sample_names(name_file, name_list):
    all_names = [name for name in name_list]
    if name_file:
        for name in name_file:
            all_names.append(name)
    out = []
    for name in all_names:
        for sub_name in name.strip().split(','):
            out.append(sub_name)
    return out


@cli_create.command('samples')
@use_common_state
@click.option('-s', '--sample-name-list', default=None, type=click.File('r'), help="A file containing a list of sample names")
@click.argument('org_name')
@click.argument('library_name')
@click.argument('sample_names', nargs=-1)
def cli_create_samples(state, sample_name_list, org_name, library_name, sample_names):
    """Create samples in the specified group.

    `sample_names` is a list of samples with one line per sample
    """
    knex = state.get_knex()
    org = Organization(knex, org_name).get()
    lib = org.sample_group(library_name, is_library=True).get()
    for sample_name in simplify_sample_names(sample_name_list, sample_names):
        sample = lib.sample(sample_name).idem()
        print(sample, file=state.outfile)


@cli_create.command('sample-with-data')
@use_common_state
@click.option('-m', '--module-name', type=click.Choice(READ_MODULE_NAMES))
@click.argument('org_name')
@click.argument('library_name')
@click.argument('sample_name')
@click.argument('data_files', nargs=-1)
def cli_create_samples(state, module_name, org_name, library_name, sample_name, data_files):
    """Create one sample in the specified group with attached raw data.

    `module_name` is the type of data being uploaded
    `org_name` is the name of the organization where the sample should be created
    `library_name` is the name of the library sample group where the sample should be created
    `sample_name` is the name for the new sample
    `data_files` is either one or two gzipped fastq files depending on the module_name
    """
    knex = state.get_knex()
    org = Organization(knex, org_name).get()
    lib = org.sample_group(library_name, is_library=True).get()
    sample = lib.sample(sample_name).idem()
    ar = sample.analysis_result(module_name).create()
    if module_name in [SINGLE_END, NANOPORE]:
        field_name=f"{module_name}::read_1::lane_1"
        r1 = ar.field(field_name).create()
        r1.upload_file(data_files[0])
    elif module_name == PAIRED_END:
        field_name=f"{module_name}::read_1::lane_1"
        r1 = ar.field(field_name).create()
        r1.upload_file(data_files[0])
        field_name=f"{module_name}::read_2::lane_1"
        r2 = ar.field(field_name).create()
        r2.upload_file(data_files[1])
    click.echo('Sample with data successfully created', err=True)


@cli_create.command('sample-ar')
@use_common_state
@click.option('-r', '--replicate')
@click.argument('names', nargs=-1)
def cli_create_sample_ar(state, replicate, names):
    """Create new sample analysis result for a sample.

    Names must be one of:
    <org name> <library name> <sample name> <new module name>
    <sample uuid> <new module name>
    
    `org name` is the name of the organization where the sample exists
    `library name` is the name of the library where the sample exists
    `sample name` is the name of the sample
    `sample uuid` is the uuid of the sample
    `new module name` module name for the created analysis result
    """
    if len(names) not in [2, 4]:
        click.echo('''
            Names must be one of:
            <org name> <library name> <sample name> <new module name>
            <sample uuid> <new module name>
        ''', err=True)
        return
    knex = state.get_knex()
    module_name = names[-1]
    if len(names) == 2:
        sample = sample_from_uuid(knex, names[0])
    else:
        org = Organization(knex, names[0]).get()
        lib = org.sample_group(names[1]).get()
        sample = lib.sample(names[2]).get()
    ar = sample.analysis_result(module_name, replicate=replicate).create()
    click.echo(f'Created: {ar}', err=True)


@cli_create.command('field')
@use_common_state
@click.argument('names', nargs=-1)
def cli_create_field(state, names):
    """Create new sample analysis result field.

    Names must be one of:
    <org name> <library name> <sample name> <module name> <new field name> <filename>
    <module uuid> <new field name> <filename>
    
    `org name` is the name of the organization where the sample exists
    `library name` is the name of the library where the sample exists
    `sample name` is the name of the sample
    `module name` module name of the target sample analysis result
    `module uuid` uuid of the target sample analysis result
    `new field name` name of the created field
    `filename` route to the selected file

    """
    if len(names) not in [3, 6]:
        click.echo('''
            Names must be one of:
            <org name> <library name> <sample name> <module name> <new field name> <filename>
            <module uuid> <new field name> <filename>
        ''', err=True)
        return
    knex = state.get_knex()
    field_name = names[-2]
    if len(names) == 3:
        ar = sample_ar_from_uuid(knex, names[0])
    else:
        org = Organization(knex, names[0]).get()
        lib = org.sample_group(names[1]).get()
        sample = lib.sample(names[2]).get()
        ar = sample.analysis_result(names[3]).get()
    arf = ar.field(field_name).create()
    click.echo(f'Created: {arf}', err=True)
    click.echo(f'Uploading file: {names[-1]}', err=True)
    arf.upload_file(names[-1], logger=lambda x: click.echo(x, err=True))
    click.echo('File successfully uploaded', err=True)



import click
import json
from geoseeq.blob_constructors import (
    sample_ar_from_uuid,
)
from .utils import use_common_state


@click.group('view')
def cli_view():
    """View an object on GeoSeeq."""
    pass


@cli_view.command('sample-result')
@use_common_state
@click.argument('uuid')
def cli_view_result(state, uuid):
    """Print a list of samples in the specified group."""
    knex = state.get_knex()
    result = sample_ar_from_uuid(knex, uuid)
    print(result)
    for field in result.get_fields():
        print('\t', field)
        for line in json.dumps(field.stored_data, indent=2).split('\n'):
            print('\t\t', line)



import json
import logging
from os import makedirs
from os.path import dirname, join

import click
import pandas as pd

from .shared_params import (
    handle_project_id,
    project_id_arg,
    sample_ids_arg,
    handle_multiple_sample_ids,
    use_common_state,
    flatten_list_of_els_and_files
)
from geoseeq.result.utils import _download_head
from geoseeq.utils import download_ftp
from geoseeq.blob_constructors import (
    sample_result_file_from_uuid,
    project_result_file_from_uuid,
)
from geoseeq.knex import GeoseeqNotFoundError
from .utils import convert_size

logger = logging.getLogger('geoseeq_api')


@click.group("download")
def cli_download():
    """Download data from GeoSeeq."""
    pass



@cli_download.command("metadata")
@use_common_state
@sample_ids_arg
def cli_download_metadata(state, sample_ids):
    """Download metadata for a set of samples as a CSV.
    
    ---

    Example Usage:

    \b
    # Download metadata for samples S1, S2, and S3 in project "My Org/My Project"
    $ geoseeq download metadata "My Org/My Project" S1 S2 S3 > metadata.csv

    \b
    # Download metadata for all samples in project "My Org/My Project"
    $ geoseeq download metadata "My Org/My Project" > metadata.csv

    \b
    # Download metadata for two samples by their UUIDs
    $ geoseeq download metadata 2b721a88-7387-4085-86df-4995d263b3f9 746424e7-2408-407e-a68d-786c7f5c5da6 > metadata.csv

    \b
    # Download metadata from a list of sample UUIDs in a file
    $ echo "2b721a88-7387-4085-86df-4995d263b3f9" > sample_ids.txt
    $ echo "746424e7-2408-407e-a68d-786c7f5c5da6" >> sample_ids.txt
    $ geoseeq download metadata sample_ids.txt > metadata.csv

    ---

    Command Arguments:

    \b
    [SAMPLE_IDS]... can be a list of sample names or IDs, files containing a list of sample names or IDs, or a mix of both.
    The first element in the list can optionally be a project ID or name.
    If a project ID is not provided, then sample ids must be UUIDs or GRNs, not names.
    If only a project ID is provided, then metadata for all samples in that project will be downloaded.

    ---
    """
    samples = handle_multiple_sample_ids(state.knex, sample_ids)
    click.echo(f"Found {len(samples)} samples.", err=True)
    metadata = {}
    for sample in samples:
        metadata[sample.name] = sample.metadata
    metadata = pd.DataFrame.from_dict(metadata, orient="index")
    metadata.to_csv(state.outfile)
    click.echo("Metadata successfully downloaded for samples.", err=True)


@cli_download.command("files")
@use_common_state
@click.option("--target-dir", default=".")
@click.option('--yes/--confirm', default=False, help='Skip confirmation prompts')
@click.option("--download/--urls-only", default=True, help="Download files or just print urls")
@click.option("--folder-type", type=click.Choice(['all', 'sample', 'project'], case_sensitive=False), default="all", help='Download files from sample folders, project folders, or both')
@click.option("--folder-name", multiple=True, help='Filter folders for names that include this string. Case insensitive.')
@click.option("--sample-name-includes", multiple=True, help='Filter samples for names that include this string. Case insensitive.')
@click.option("--file-name", multiple=True, help="Filter files for names that include this string. Case insensitive.")
@click.option("--extension", multiple=True, help="Only download files with this extension. e.g. 'fastq.gz', 'bam', 'csv'")
@click.option("--with-versions/--without-versions", default=False, help="Download all versions of a file, not just the latest")
@project_id_arg
@sample_ids_arg
def cli_download_files(
    state,
    sample_name_includes,
    target_dir,
    yes,
    folder_type,
    folder_name,
    file_name,
    extension,
    with_versions,
    download,
    project_id,
    sample_ids,
):
    """Download files from a GeoSeeq project.
    
    This command will download multiple files from a GeoSeeq project. You can filter
    files by file extension, folder name, sample name, file name, and file extension.
    You can also choose to download all versions of a file, not just the latest.

    ---

    Example Usage:

    \b
    # Download all fastq files from all samples in "My Org/My Project"
    $ geoseeq download files "My Org/My Project" --extension fastq.gz --folder-type sample

    \b
    # Download fastq files from the MetaSUB Consortium CSD16 project that have been cleaned
    # e.g. "https://portal.geoseeq.com/samples/9c60aa67-eb3d-4b02-9c77-94e22361b2f3/analysis-results/b4ae15d2-37b9-448b-9946-3d716826eaa8"
    $ geoseeq download files "MetaSUB Consortium/CSD16" \\
        --folder-type sample `# Only download files from sample folders, not project folders` \\
        --folder-name "clean_reads"  # created by MetaSUB , not all projects will have this folder

    \b
    # Download a table of taxonomic abundances from the MetaSUB Consortium CSD17 project
    # produced by the GeoSeeq Search tool
    $ geoseeq download files "MetaSUB Consortium/CSD17" --folder-type project --folder-name "GeoSeeq Search" --file-name "Taxa Table"

    \b
    # Download assembly contigs from two samples in the MetaSUB Consortium CSD16 project
    $ geoseeq download files "MetaSUB Consortium/CSD16" `# specify the project` \ 
        haib17CEM4890_H2NYMCCXY_SL254769 haib17CEM4890_H2NYMCCXY_SL254773 `# specify the samples by name` \ 
        --folder-type sample --extension '.contigs.fasta' # filter for contig files

    ---

    Command Arguments:

    [PROJECT_ID] Can be a project UUID, GeoSeeq Resource Number (GRN), or an
    organization name and project name separated by a slash.
    
    \b
    [SAMPLE_IDS]... can be a list of sample names or IDs, files containing a list of sample names or IDs, or a mix of both.
    ---
    """
    knex = state.get_knex()
    proj = handle_project_id(knex, project_id)
    samples = []
    if sample_ids:
        logger.info(f"Fetching info for {len(sample_ids)} samples.")
        samples = handle_multiple_sample_ids(knex, sample_ids, proj=proj)

    data = {
        "sample_uuids": [s.uuid for s in samples],
        "sample_names": sample_name_includes,
        "folder_type": folder_type,
        "folder_names": folder_name,
        "file_names": file_name,
        "extensions": extension,
        "with_versions": with_versions
    }
    url = f"sample_groups/{proj.uuid}/download"
    response = knex.post(url, data)

    if not download:
        data = json.dumps(response["links"])
        print(data, file=state.outfile)

    else:
        files_size = convert_size(response['file_size_bytes'])
        no_size_info = f"No size info was found for {response['no_size_info_count']} files." if response['no_size_info_count'] else ""
        click.echo(f"Found {len(response['links'])} files to download with total size of {files_size}. {no_size_info}\n")
        for fname, url in response["links"].items():
            clean_url = url.split("?")[0]
            click.echo(f'{clean_url} -> {target_dir}/{fname}')
        if not yes:
            click.confirm('Do you want to download these files?', abort=True)

        for fname, url in response["links"].items():
            click.echo(f"Downloading file {fname}")
            file_path = join(target_dir, fname)
            makedirs(dirname(file_path), exist_ok=True)
            if url.startswith("ftp"):
                download_ftp(url, file_path)
            else:
                _download_head(url, file_path)


@cli_download.command("ids")
@use_common_state
@click.option("--target-dir", default=".")
@click.option('--yes/--confirm', default=False, help='Skip confirmation prompts')
@click.option("--download/--urls-only", default=True, help="Download files or just print urls")
@click.argument("ids", nargs=-1)
def cli_download_ids(state, target_dir, yes, download, ids):
    """Download a files from GeoSeeq based on their UUID or GeoSeeq Resource Number (GRN).

    This command downloads files directly based on their ID. This is used for "manual"
    downloads of relatively small amounts of data. For bulk downloads, use the
    `geoseeq download files` command.

    ---

    Example Usage:

    # Download a single file
    $ geoseeq download ids 9c60aa67-eb3d-4b02-9c77-94e22361b2f3

    # Download multiple files
    $ geoseeq download ids 9c60aa67-eb3d-4b02-9c77-94e22361b2f3 9c60aa67-eb3d-4b02-9c77-94e22361b2f3

    ---

    Command Arguments:

    [IDS]... can be a list of sample names or IDs, files containing a list of sample names or IDs, or a mix of both.

    ---
    """
    result_file_ids = flatten_list_of_els_and_files(ids)
    knex = state.get_knex()
    result_files = []
    for result_id in result_file_ids:
        result_uuid = result_id.split(':')[-1]
        # we guess that this is a sample file, TODO: use GRN if available
        try:
            result_file = sample_result_file_from_uuid(knex, result_uuid)
        except GeoseeqNotFoundError:
            result_file = project_result_file_from_uuid(knex, result_uuid)
        result_files.append(result_file)

    if not download:
        for result_file in result_files:
            print(result_file.get_download_url(), file=state.outfile)
        return
    
    for result_file in result_files:
        click.echo(f"{result_file} -> {target_dir}/{result_file.get_referenced_filename()}")
    if not yes:
        click.confirm('Do you want to download these files?', abort=True)

    for result_file in result_files:
        click.echo(f"Downloading file {result_file.get_referenced_filename()}")
        file_path = join(target_dir, result_file.get_referenced_filename())
        makedirs(dirname(file_path), exist_ok=True)
        result_file.download(file_path)



from .id_handlers import (
    handle_project_id,
    flatten_list_of_els_and_files,
    handle_multiple_sample_ids,
    handle_folder_id,
)
from .opts_and_args import *
from .common_state import use_common_state


import logging
import json

import click
import pandas as pd
import requests
from os.path import basename
from geoseeq.knex import GeoseeqNotFoundError
from multiprocessing import Pool, current_process

from geoseeq import Organization
from geoseeq.cli.constants import *
from geoseeq.cli.fastq_utils import group_paired_end_paths, upload_fastq_pair, upload_single_fastq

from geoseeq.cli.shared_params import (
    use_common_state,
    yes_option,
    private_option,
    link_option,
    folder_id_arg,
    handle_folder_id,
    overwrite_option,
    project_id_arg,
    handle_project_id,
)

logger = logging.getLogger('geoseeq_api')


@click.command('files')
@use_common_state
@yes_option
@private_option
@link_option
@click.option('-n', '--geoseeq-file-name', default=None, multiple=True,
              help='Specify a different name for the file on GeoSeeq than the local file name')
@folder_id_arg
@click.argument('file_paths', type=click.Path(exists=True), nargs=-1)
def cli_upload_file(state, yes, private, link_type, geoseeq_file_name, folder_id, file_paths):
    """Upload files to GeoSeeq.

    This command uploads files to either a sample or project on GeoSeeq. It can be used to upload
    multiple files to the same folder at once.
    
    ---

    Example Usage:

    \b
    # Upload file.txt to a sample
    $ geoseeq upload files "My Org/My Project/My Sample/My Folder" /path/to/file.txt

    \b
    # Upload file.txt to a sample but name it "My File" on GeoSeeq
    $ geoseeq upload files "My Org/My Project/My Sample/My Folder" /path/to/file.txt -n "My File"

    \b
    # Upload file.txt to a project
    $ geoseeq upload files "My Org/My Project/My Folder" /path/to/file.txt

    \b
    # Upload multiple files to a project
    $ geoseeq upload files "My Org/My Project/My Folder" /path/to/file1.txt /path/to/file2.txt

    \b
    # Upload multiple files to a project but name them differently on GeoSeeq
    $ geoseeq upload files "My Org/My Project/My Folder" /path/to/file1.txt /path/to/file2.txt -n "File 1" -n "File 2"

    ---

    Command Arguments:

    [FOLDER_ID] Can be a folder UUID, GeoSeeq Resource Number (GRN), or an
    names for an org, project, sample, folder separated by a slash. Can exclude
    the sample name if the folder is for a project.

    [FILE_PATHS]... One or more paths to files on your local machine.

    ---
    """
    knex = state.get_knex()
    result_folder = handle_folder_id(knex, folder_id, yes=yes, private=private, create=True)
    if geoseeq_file_name:
        if len(geoseeq_file_name) != len(file_paths):
            raise click.UsageError('Number of --geoseeq-file-name arguments must match number of file_paths')
        name_pairs = zip(geoseeq_file_name, file_paths)
    else:
        name_pairs = zip([basename(fp) for fp in file_paths], file_paths)
    for geoseeq_file_name, file_path in name_pairs:
        field = result_folder.result_file(geoseeq_file_name).idem()
        if link_type == 'upload':
            field.upload_file(file_path)
        else:
            field.link_file(link_type, file_path)



@click.command('metadata')
@use_common_state
@overwrite_option
@yes_option
@private_option
@click.option('--create/--no-create', default=False, help='Create samples if they have metadata but do not exist on GeoSeeq.')
@click.option('--index-col', default=0)
@click.option('--encoding', default='utf_8')
@project_id_arg
@click.argument('table', type=click.File('rb'))
def cli_metadata(state, overwrite, yes, private, create, index_col, encoding, project_id, table):
    """Upload sample metadata to a project on GeoSeeq.
    
    This command takes a CSV file with one row per sample and one column per metadata field.
    
    ---

    Example Usage:

    \b
    # Upload metadata from a metadata.csv to a project
    $ cat metadata.csv
        sample_name,collection_date,location
        sample1,2020-01-01,USA
        sample2,2020-01-02,USA
        sample3,2020-01-03,CAN
    $ geoseeq upload metadata "My Org/My Project" metadata.csv

    \b
    # Modify metadata for existing samples
    $ cat fixed_metadata.csv
        sample_name,collection_date,location
        sample1,2020-01-01,CAN
    $ geoseeq upload metadata --overwrite "My Org/My Project" fixed_metadata.csv

    ---

    Command Arguments:

    [PROJECT_ID] Can be a project UUID, GeoSeeq Resource Number (GRN), or an
    organization name and project name separated by a slash.

    [TABLE] A CSV file with one row per sample and one column per metadata field.

    ---
    """
    knex = state.get_knex()
    proj = handle_project_id(knex, project_id, yes, private)
    tbl = pd.read_csv(table, index_col=index_col, encoding=encoding)
    samples = []
    plan = {
        'overwrite': 0,
        'new': 0,
        'no_change': 0,
        'no_metadata': 0,
    }
    for sample_name, row in tbl.iterrows():
        sample = proj.sample(sample_name)
        if create:
            sample = sample.idem()
        else:
            sample = sample.get()
        new_meta = json.loads(json.dumps(row.dropna().to_dict()))
        if new_meta:
            if sample.metadata == new_meta:
                plan['no_change'] += 1
            elif overwrite and sample.metadata:
                plan['overwrite'] += 1
                samples.append((sample, new_meta))
            elif not sample.metadata:
                plan['new'] += 1
                samples.append((sample, new_meta))
        else:
            plan['no_metadata'] += 1


    
    if not yes:
        click.echo('Plan:')
        click.echo(f'{plan["overwrite"]} samples will have their metadata overwritten')
        click.echo(f'{plan["new"]} samples will have metadata added')
        click.echo(f'{plan["no_change"]} samples have metadata that already matches the new metadata to upload')
        click.echo(f'{plan["no_metadata"]} samples have no metadata to upload')
        click.confirm('Continue?', abort=True)

    for sample, new_meta in samples:
        if overwrite or (not sample.metadata):
            sample.metadata = new_meta
            sample.idem()
    click.echo(f'Wrote metadata for {len(samples)} samples')



from geoseeq.constants import FASTQ_MODULE_NAMES


class SampleBioInfoFolder:
    """Abstract class that adds bioinformatic functionality to a SampleResultFolder."""

    @property
    def is_fastq(self):
        return self.module_name in FASTQ_MODULE_NAMES
    
    @property
    def is_short_read(self):
        return self.module_name.startswith('short_read')
    
    @property
    def is_paired_end(self):
        return self.module_name.endswith('paired_end')
    
    def read_file(self, read_id='read_1::lane_1'):
        assert self.is_fastq
        if 'read_2' in read_id: # paired end
            assert self.is_paired_end
        file_name = f'{self.module_name}::{read_id}'
        result_file = self.result_file(file_name)
        return result_file
    
    def read_1(self, lane_id='lane_1'):
        assert self.is_fastq
        file_name = f'{self.module_name}::read_1::{lane_id}'
        result_file = self.result_file(file_name)
        return result_file
    
    def read_2(self, lane_id='lane_1'):
        assert self.is_fastq
        assert self.is_paired_end
        file_name = f'{self.module_name}::read_2::{lane_id}'
        result_file = self.result_file(file_name)
        return result_file
    
    @classmethod
    def fastq_folder(cls, sample, read_length='short_read', paired_end=False, long_read_type='nanopore'):
        if read_length == 'short_read':
            if paired_end:
                module_name = 'short_read::paired_end'
            else:
                module_name = 'short_read::single_end'
        elif read_length == 'long_read':
            module_name = 'long_read::{long_read_type}'
        else:
            raise ValueError(f'Invalid read_length: {read_length}')
        return sample.result_folder(module_name)
        



import urllib.request
import logging
from os.path import basename, getsize, join
from pathlib import Path
from tempfile import NamedTemporaryFile

from geoseeq.utils import download_ftp

from .utils import  _download_head

logger = logging.getLogger("geoseeq_api")  # Same name as calling module


class ResultFileDownload:
    """Abstract class that handles download methods for result files."""

    def get_download_url(self):
        """Return a URL that can be used to download the file for this result."""
        blob_type = self.stored_data.get("__type__", "").lower()
        if blob_type not in ["s3", "sra"]:
            raise TypeError("Cannot fetch a file for a BLOB type result field.")
        if blob_type == "s3":
            try:
                url = self.stored_data["presigned_url"]
            except KeyError:
                url = self.stored_data["uri"]
            if url.startswith("s3://"):
                url = self.stored_data["endpoint_url"] + "/" + url[5:]
            return url
        elif blob_type == "sra":
            url = self.stored_data["url"]
            return url

    def download_file(self, filename=None, cache=True, head=None):
        """Return a local filepath to the file this result points to."""
        if not filename:
            self._temp_filename = True
            myfile = NamedTemporaryFile(delete=False)
            myfile.close()
            filename = myfile.name
        blob_type = self.stored_data.get("__type__", "").lower()
        if cache and self._cached_filename:
            return self._cached_filename
        if blob_type == "s3":
            return self._download_s3(filename, cache, head=head)
        elif blob_type == "sra":
            return self._download_sra(filename, cache)
        elif blob_type == "ftp":
            return self._download_ftp(filename, cache)
        elif blob_type == "azure":
            return self._download_azure(filename, cache, head=head)
        else:
            raise TypeError("Cannot fetch a file for a BLOB type result field.")

    def _download_s3(self, filename, cache, head=None):
        logger.info(f"Downloading S3 file to {filename}")
        try:
            url = self.stored_data["presigned_url"]
        except KeyError:
            key = 'uri' if 'uri' in self.stored_data else 'url'
            url = self.stored_data[key]
        if url.startswith("s3://"):
            url = self.stored_data["endpoint_url"] + "/" + url[5:]
        _download_head(url, filename, head=head) 
        if cache:
            self._cached_filename = filename
        return filename

    def _download_azure(self, filename, cache, head=None):
        logger.info(f"Downloading Azure file to {filename}")
        try:
            url = self.stored_data["presigned_url"]
        except KeyError:
            key = 'uri' if 'uri' in self.stored_data else 'url'
            url = self.stored_data[key]
        _download_head(url, filename, head=head)
        if cache:
            self._cached_filename = filename
        return filename

    def _download_sra(self, filename, cache):
        return self._download_generic_url(filename, cache)

    def _download_ftp(self, filename, cache, head=None):
        logger.info(f"Downloading FTP file to {filename}")
        key = 'url' if 'url' in self.stored_data else 'uri'
        download_ftp(self.stored_data[key], filename, head=head)
        return filename

    def _download_generic_url(self, filename, cache):
        logger.info(f"Downloading generic URL file to {filename}")
        key = 'url' if 'url' in self.stored_data else 'uri'
        url = self.stored_data[key]
        urllib.request.urlretrieve(url, filename)
        if cache:
            self._cached_filename = filename
        return filename
